import unittest
import numpy as np
import pandas as pd
import logging

from ..timeseries import TimeSeries
from ..metrics import metrics


class MetricsTestCase(unittest.TestCase):

    pd_train = pd.Series(np.sin(np.pi * np.arange(31) / 4) + 1, index=pd.date_range('20121201', '20121231'))
    pd_series1 = pd.Series(range(10), index=pd.date_range('20130101', '20130110'))
    pd_series2 = pd.Series(np.random.rand(10) * 10 + 1, index=pd.date_range('20130101', '20130110'))
    pd_series3 = pd.Series(np.sin(np.pi * np.arange(20) / 4) + 1, index=pd.date_range('20130101', '20130120'))
    series_train = TimeSeries.from_series(pd_train)
    series1: TimeSeries = TimeSeries.from_series(pd_series1)
    pd_series1[:] = pd_series1.mean()
    series0: TimeSeries = TimeSeries.from_series(pd_series1)
    series2: TimeSeries = TimeSeries.from_series(pd_series2)
    series3: TimeSeries = TimeSeries.from_series(pd_series3)
    series12: TimeSeries = series1.stack(series2)
    series21: TimeSeries = series2.stack(series1)
    series1b = TimeSeries.from_times_and_values(pd.date_range('20130111', '20130120'), series1.values())
    series2b = TimeSeries.from_times_and_values(pd.date_range('20130111', '20130120'), series2.values())

    @classmethod
    def setUpClass(cls):
        logging.disable(logging.CRITICAL)

    def test_zero(self):
        with self.assertRaises(ValueError):
            metrics.mape(self.series1, self.series1)

        with self.assertRaises(ValueError):
            metrics.smape(self.series1, self.series1)

        with self.assertRaises(ValueError):
            metrics.mape(self.series12, self.series12)

        with self.assertRaises(ValueError):
            metrics.smape(self.series12, self.series12)

        with self.assertRaises(ValueError):
            metrics.ope(self.series1 - self.series1.pd_series().mean(), self.series1 - self.series1.pd_series().mean())

    def test_same(self):
        self.assertEqual(metrics.mape(self.series1 + 1, self.series1 + 1), 0)
        self.assertEqual(metrics.smape(self.series1 + 1, self.series1 + 1), 0)
        self.assertEqual(metrics.mase(self.series1 + 1, self.series1 + 1, self.series_train, 1), 0)
        self.assertEqual(metrics.marre(self.series1 + 1, self.series1 + 1), 0)
        self.assertEqual(metrics.r2_score(self.series1 + 1, self.series1 + 1), 1)
        self.assertEqual(metrics.ope(self.series1 + 1, self.series1 + 1), 0)

    def helper_test_shape_equality(self, metric):
        self.assertAlmostEqual(metric(self.series12, self.series21),
                               metric(self.series1.append(self.series2b), self.series2.append(self.series1b)))

    def helper_test_multivariate_duplication_equality(self, metric, **kwargs):
        series11 = self.series1.stack(self.series1) + 1
        series22 = self.series2.stack(self.series2)
        series33 = self.series3.stack(self.series3)
        self.assertAlmostEqual(metric(series11, series22, **kwargs), metric(self.series1 + 1, self.series2, **kwargs))
        self.assertAlmostEqual(metric(series11, series33, **kwargs), metric(self.series1 + 1, self.series3, **kwargs))
        self.assertAlmostEqual(metric(series22, series33, **kwargs), metric(self.series2, self.series3, **kwargs))
        self.assertAlmostEqual(metric(series22, series33, reduction=(lambda x: x[0]), **kwargs),
                               metric(self.series2, self.series3, reduction=(lambda x: x[0]), **kwargs))

    def test_r2(self):
        from sklearn.metrics import r2_score
        self.assertEqual(metrics.r2_score(self.series1, self.series0), 0)
        self.assertEqual(metrics.r2_score(self.series1, self.series2),
                         r2_score(self.series1.values(), self.series2.values()))
        self.helper_test_multivariate_duplication_equality(metrics.r2_score)

    def test_marre(self):
        self.assertAlmostEqual(metrics.marre(self.series1, self.series2),
                               metrics.marre(self.series1 + 100, self.series2 + 100))
        self.helper_test_multivariate_duplication_equality(metrics.marre)

    def test_season(self):
        with self.assertRaises(ValueError):
            metrics.mase(self.series3, self.series3 * 1.3, self.series_train, 8)

    def test_mse(self):
        self.helper_test_shape_equality(metrics.mse)

    def test_mae(self):
        self.helper_test_shape_equality(metrics.mae)

    def test_rmse(self):
        self.helper_test_multivariate_duplication_equality(metrics.rmse)

        self.assertAlmostEqual(metrics.rmse(self.series1.append(self.series2b), self.series2.append(self.series1b)),
                               metrics.mse(self.series12, self.series21, reduction=(lambda x: np.sqrt(np.mean(x)))))

    def test_rmsle(self):
        self.helper_test_multivariate_duplication_equality(metrics.rmsle)

    def test_coefficient_of_variation(self):
        self.helper_test_multivariate_duplication_equality(metrics.coefficient_of_variation)

    def test_mape(self):
        self.helper_test_multivariate_duplication_equality(metrics.mape)

    def test_smape(self):
        self.helper_test_multivariate_duplication_equality(metrics.smape)

    def test_mase(self):
        self.helper_test_multivariate_duplication_equality(metrics.mase, insample=self.series_train)

        with self.assertRaises(ValueError):
            metrics.mase(self.series1, self.series2, self.series3, 1)

    def test_ope(self):
        self.helper_test_multivariate_duplication_equality(metrics.ope)

    def test_r2_score(self):
        self.helper_test_multivariate_duplication_equality(metrics.r2_score)

    def test_metrics_arguments(self):
        series00 = self.series0.stack(self.series0)
        series11 = self.series1.stack(self.series1)
        self.assertEqual(metrics.r2_score(series11, series00, np.mean), 0)
        self.assertEqual(metrics.r2_score(series11, series00, reduction=np.mean), 0)
        self.assertEqual(metrics.r2_score(series11, series2=series00, reduction=np.mean), 0)
        self.assertEqual(metrics.r2_score(series00, series1=series11, reduction=np.mean), 0)
        self.assertEqual(metrics.r2_score(np.mean, series2=series00, series1=series11), 0)
        self.assertEqual(metrics.r2_score(series00, np.mean, series1=series11), 0)
        self.assertEqual(metrics.r2_score(series11, np.mean, series2=series00), 0)



"""
Utils for filling missing values
--------------------------------
"""

from ..timeseries import TimeSeries


def na_ratio(ts: TimeSeries) -> float:
    """
    Computes the ratio of missing values

    Parameters
    ----------
    ts
        The time series to compute ratio on
    Returns
    -------
    float
        The ratio of missing values
    """

    return ts.pd_dataframe().isnull().sum().mean() / len(ts)


def fillna(ts: TimeSeries, fill: float = 0) -> TimeSeries:
    """
    Fills the missing values of `ts` with only the value provided (default zeroes).

    Parameters
    ----------
    ts
        The TimeSeries to check for missing values.
    fill
        The value used to replace the missing values.

    Returns
    -------
    TimeSeries
        A TimeSeries, `ts` with all missing values set to `fill`.
    """

    return TimeSeries.from_times_and_values(ts.time_index(), ts.pd_dataframe().fillna(value=fill), ts.freq())


def auto_fillna(ts: TimeSeries,
                **interpolate_kwargs) -> TimeSeries:
    """
    This function fills the missing value in the TimeSeries `ts`,
    using the `pandas.Dataframe.interpolate()` method.

    Parameters
    ----------
    ts
        The time series
    interpolate_kwargs
        Keyword arguments  `pandas.Dataframe.interpolate()`.
        See `the documentation
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.interpolate.html>`_
        for the list of supported parameters.
    Returns
    -------
    TimeSeries
        A new TimeSeries with all missing values filled according to the rules above.
    """

    ts_temp = ts.pd_dataframe()

    # pandas interpolate wrapper, with chosen `method`
    if 'limit_direction' not in interpolate_kwargs:
        interpolate_kwargs['limit_direction'] = 'both'
    interpolate_kwargs['inplace'] = True
    ts_temp.interpolate(**interpolate_kwargs)

    return TimeSeries.from_times_and_values(ts.time_index(), ts_temp.values, ts.freq())



import unittest
import pandas as pd

from ..utils import retain_period_common_to_all, _with_sanity_checks
from ..timeseries import TimeSeries


class UtilsTestCase(unittest.TestCase):

    def test_retain_period_common_to_all(self):
        seriesA = TimeSeries.from_times_and_values(pd.date_range('20000101', '20000110'), range(10))
        seriesB = TimeSeries.from_times_and_values(pd.date_range('20000103', '20000108'), range(6))
        seriesC = TimeSeries.from_times_and_values(pd.date_range('20000104', '20000112'), range(9))
        seriesC = seriesC.stack(seriesC)

        common_series_list = retain_period_common_to_all([seriesA, seriesB, seriesC])

        # test start and end dates
        for common_series in common_series_list:
            self.assertEqual(common_series.start_time(), pd.Timestamp('20000104'))
            self.assertEqual(common_series.end_time(), pd.Timestamp('20000108'))

        # test widths
        self.assertEqual(common_series_list[0].width, 1)
        self.assertEqual(common_series_list[1].width, 1)
        self.assertEqual(common_series_list[2].width, 2)

    def test_sanity_check_example(self):
        class Model:
            def _sanity_check(self, *args, **kwargs):
                if kwargs['b'] != kwargs['c']:
                    raise(ValueError("b and c must be equal"))

            @_with_sanity_checks("_sanity_check")
            def fit(self, a, b=0, c=0):
                pass

        m = Model()

        # b != c should raise error
        with self.assertRaises(ValueError):
            m.fit(5, b=3, c=2)

        # b == c should not raise error
        m.fit(5, b=2, c=2)



import logging
import shutil
import unittest

import numpy as np
import pandas as pd

from ..timeseries import TimeSeries
from ..utils import timeseries_generation as tg
from ..metrics import mape
from ..models import NaiveSeasonal, ExponentialSmoothing, ARIMA, Theta, FourTheta, FFT
from .. import SeasonalityMode, TrendMode, ModelMode
from ..logging import get_logger

logger = get_logger(__name__)

# (forecasting models, maximum error) tuples
models = [
        (ExponentialSmoothing(), 4.8),
        (ARIMA(0, 1, 1), 17.1),
        (ARIMA(1, 1, 1), 14.2),
        (Theta(), 11.3),
        (Theta(1), 20.2),
        (Theta(-1), 9.8),
        (FourTheta(1), 20.2),
        (FourTheta(-1), 9.8),
        (FourTheta(trend_mode=TrendMode.EXPONENTIAL), 5.5),
        (FourTheta(model_mode=ModelMode.MULTIPLICATIVE), 11.4),
        (FourTheta(season_mode=SeasonalityMode.ADDITIVE), 14.2),
        (FFT(trend='poly'), 11.4),
        (NaiveSeasonal(), 32.4),
    ]

try:
    from ..models import Prophet
    models.append((Prophet(), 13.5))
except ImportError:
    logger.warning('Prophet not installed - will be skipping Prophet tests')

try:
    from ..models import AutoARIMA
    models.append((AutoARIMA(), 13.7))
except ImportError:
    logger.warning('pmdarima not installed - will be skipping AutoARIMA tests')

try:
    from ..models import RNNModel, TCNModel
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch not installed - will be skipping Torch models tests')
    TORCH_AVAILABLE = False


class ForecastingModelsTestCase(unittest.TestCase):

    # forecasting horizon used in runnability tests
    forecasting_horizon = 5

    # dummy timeseries for runnability tests
    np.random.seed(1)
    ts_gaussian = tg.gaussian_timeseries(length=100, mean=50)

    # real timeseries for functionality tests
    df = pd.read_csv('examples/AirPassengers.csv', delimiter=",")
    ts_passengers = TimeSeries.from_dataframe(df, 'Month', ['#Passengers'])
    ts_pass_train, ts_pass_val = ts_passengers.split_after(pd.Timestamp('19570101'))

    @classmethod
    def setUpClass(cls):
        logging.disable(logging.CRITICAL)

    @classmethod
    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def tearDownClass(cls):
        shutil.rmtree('.darts')

    def test_models_runnability(self):
        for model, _ in models:
            model.fit(self.ts_gaussian)
            prediction = model.predict(self.forecasting_horizon)
            self.assertTrue(len(prediction) == self.forecasting_horizon)

    def test_models_performance(self):
        # for every model, check whether its errors do not exceed the given bounds
        for model, max_mape in models:
            model.fit(self.ts_pass_train)
            prediction = model.predict(len(self.ts_pass_val))
            current_mape = mape(prediction, self.ts_pass_val)
            self.assertTrue(current_mape < max_mape, "{} model exceeded the maximum MAPE of {}."
                            "with a MAPE of {}".format(str(model), max_mape, current_mape))

    def test_multivariate_input(self):
        es_model = ExponentialSmoothing()
        ts_passengers_enhanced = self.ts_passengers.add_datetime_attribute('month')
        with self.assertRaises(AssertionError):
            es_model.fit(ts_passengers_enhanced)
        es_model.fit(ts_passengers_enhanced["#Passengers"])
        with self.assertRaises(KeyError):
            es_model.fit(ts_passengers_enhanced["2"])

        if TORCH_AVAILABLE:
            tcn_model = TCNModel(n_epochs=1, input_size=2)
            with self.assertRaises(ValueError):
                tcn_model.fit(ts_passengers_enhanced)
            tcn_model.fit(ts_passengers_enhanced, ts_passengers_enhanced["Month"])
            with self.assertRaises(KeyError):
                tcn_model.fit(ts_passengers_enhanced, ts_passengers_enhanced["2"])
            tcn_model = TCNModel(n_epochs=1, input_size=2, output_size=2)
            with self.assertRaises(KeyError):
                tcn_model.fit(ts_passengers_enhanced, ts_passengers_enhanced[["#Passengers", "2"]])
            tcn_model.fit(ts_passengers_enhanced, ts_passengers_enhanced[["#Passengers", "Month"]])



"""
Metrics
-------

Some metrics to compare time series.
"""

import numpy as np
from typing import Tuple
from ..timeseries import TimeSeries
from ..utils.statistics import check_seasonality
from ..logging import raise_if_not, get_logger
from warnings import warn
from typing import Optional, Callable
from inspect import signature

logger = get_logger(__name__)


def multivariate_support(func):
    """
    This decorator transforms a metric function that takes as input two univariate TimeSeries instances
    into a function that takes two equally-sized multivariate TimeSeries instances, computes the pairwise univariate
    metrics for components with the same indices, and returns a float value that is computed as a function of all the
    univariate metrics using a `reduction` subroutine passed as argument to the metric function.
    """

    def wrapper_multivariate_support(*args, **kwargs):
        series1 = kwargs['series1'] if 'series1' in kwargs else args[0]
        series2 = kwargs['series2'] if 'series2' in kwargs else args[0] if 'series1' in kwargs else args[1]

        raise_if_not(series1.width == series2.width, "The two TimeSeries instances must have the same width.", logger)

        num_series_in_args = int('series1' not in kwargs) + int('series2' not in kwargs)
        kwargs.pop('series1', 0)
        kwargs.pop('series2', 0)
        value_list = []
        for i in range(series1.width):
            value_list.append(func(series1.univariate_component(i), series2.univariate_component(i),
                              *args[num_series_in_args:], **kwargs))
        if 'reduction' in kwargs:
            return kwargs['reduction'](value_list)
        else:
            return signature(func).parameters['reduction'].default(value_list)
    return wrapper_multivariate_support


def _get_values_or_raise(series_a: TimeSeries,
                         series_b: TimeSeries,
                         intersect: bool) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns the numpy values of two time series. If intersect is true, considers only their time intersection.
    Raises a ValueError if the two time series (or their intersection) do not have the same time index.
    """

    raise_if_not(series_a.width == series_b.width, " The two time series must have the same number of components",
                 logger)

    series_a_common = series_a.slice_intersect(series_b) if intersect else series_a
    series_b_common = series_b.slice_intersect(series_a) if intersect else series_b

    raise_if_not(series_a_common.has_same_time_as(series_b_common), 'The two time series (or their intersection) '
                                                                    'must have the same time index.'
                                                                    '\nFirst series: {}\nSecond series: {}'.format(
                                                                    series_a.time_index(), series_b.time_index()),
                 logger)

    return series_a_common.univariate_values(), series_b_common.univariate_values()


@multivariate_support
def mae(series1: TimeSeries,
        series2: TimeSeries,
        intersect: bool = True,
        reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Mean Absolute Error (MAE).

    For two time series :math:`y^1` and :math:`y^2` of length :math:`T`, it is computed as

    .. math:: \\frac{1}{T}\\sum_{t=1}^T{(|y^1_t - y^2_t|)}.

    Parameters
    ----------
    series1
        The first time series
    series2
        The second time series
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Mean Absolute Error (MAE)
    """

    y1, y2 = _get_values_or_raise(series1, series2, intersect)
    return np.mean(np.abs(y1 - y2))


@multivariate_support
def mse(series1: TimeSeries,
        series2: TimeSeries,
        intersect: bool = True,
        reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Mean Squared Error (MSE).

    For two time series :math:`y^1` and :math:`y^2` of length :math:`T`, it is computed as

    .. math:: \\frac{1}{T}\\sum_{t=1}^T{(y^1_t - y^2_t)^2}.

    Parameters
    ----------
    series1
        The first time series
    series2
        The second time series
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Mean Squared Error (MSE)
    """

    y_true, y_pred = _get_values_or_raise(series1, series2, intersect)
    return np.mean((y_true - y_pred)**2)


@multivariate_support
def rmse(series1: TimeSeries,
         series2: TimeSeries,
         intersect: bool = True,
         reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Root Mean Squared Error (RMSE).

    For two time series :math:`y^1` and :math:`y^2` of length :math:`T`, it is computed as

    .. math:: \\sqrt{\\frac{1}{T}\\sum_{t=1}^T{(y^1_t - y^2_t)^2}}.

    Parameters
    ----------
    series1
        The first time series
    series2
        The second time series
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Root Mean Squared Error (RMSE)
    """
    return np.sqrt(mse(series1, series2, intersect))


@multivariate_support
def rmsle(series1: TimeSeries,
          series2: TimeSeries,
          intersect: bool = True,
          reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Root Mean Squared Log Error (RMSLE).

    For two time series :math:`y^1` and :math:`y^2` of length :math:`T`, it is computed as

    .. math:: \\sqrt{\\frac{1}{T}\\sum_{t=1}^T{\\left(\\log{(y^1_t + 1)} - \\log{(y^2_t + 1)}\\right)^2}},

    using the natural logarithm.

    Parameters
    ----------
    series1
        The first time series
    series2
        The second time series
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Root Mean Squared Log Error (RMSLE)
    """

    y1, y2 = _get_values_or_raise(series1, series2, intersect)
    y1, y2 = np.log(y1 + 1), np.log(y2 + 1)
    return np.sqrt(np.mean((y1 - y2)**2))


@multivariate_support
def coefficient_of_variation(actual_series: TimeSeries,
                             pred_series: TimeSeries,
                             intersect: bool = True,
                             reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Coefficient of Variation (percentage).

    Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`,
    it is a percentage value, computed as

    .. math:: 100 \\cdot \\text{RMSE}(y_t, \\hat{y}_t) / \\bar{y_t},

    where :math:`\\text{RMSE}()` denotes the root mean squred error, and
    :math:`\\bar{y_t}` is the average of :math:`y_t`.

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Coefficient of Variation
    """

    return 100 * rmse(actual_series, pred_series, intersect) / actual_series.mean().mean()


@multivariate_support
def mape(actual_series: TimeSeries,
         pred_series: TimeSeries,
         intersect: bool = True,
         reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Mean Absolute Percentage Error (MAPE).

    Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
    both of length :math:`T`, it is a percentage value computed as

    .. math:: 100 \\cdot \\frac{1}{T} \\sum_{t=1}^{T}{\\left| \\frac{y_t - \\hat{y}_t}{y_t} \\right|}.

    Note that it will raise a `ValueError` if :math:`y_t = 0` for some :math:`t`. Consider using
    the Mean Absolute Scaled Error (MASE) in these cases.

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Raises
    ------
    ValueError
        If the actual series contains some zeros.

    Returns
    -------
    float
        The Mean Absolute Percentage Error (MAPE)
    """

    y_true, y_hat = _get_values_or_raise(actual_series, pred_series, intersect)
    raise_if_not((y_true != 0).all(), 'The actual series must be strictly positive to compute the MAPE.', logger)
    return 100. * np.mean(np.abs((y_true - y_hat) / y_true))


@multivariate_support
def smape(actual_series: TimeSeries,
          pred_series: TimeSeries,
          intersect: bool = True,
          reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ symmetric Mean Absolute Percentage Error (sMAPE).

    Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
    both of length :math:`T`, it is a percentage value computed as

    .. math::
        200 \\cdot \\frac{1}{T}
        \\sum_{t=1}^{T}{\\frac{\\left| y_t - \\hat{y}_t \\right|}{\\left| y_t \\right| + \\left| \\hat{y}_t \\right|} }.

    Note that it will raise a `ValueError` if :math:`\\left| y_t \\right| + \\left| \\hat{y}_t \\right| = 0`
     for some :math:`t`. Consider using the Mean Absolute Scaled Error (MASE) in these cases.

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Raises
    ------
    ValueError
        If the actual series and the pred series contains some zeros at the same time index.

    Returns
    -------
    float
        The symmetric Mean Absolute Percentage Error (sMAPE)
    """

    y_true, y_hat = _get_values_or_raise(actual_series, pred_series, intersect)
    raise_if_not(np.logical_or(y_true != 0, y_hat != 0).all(),
                 'The actual series must be strictly positive to compute the sMAPE.', logger)
    return 200. * np.mean(np.abs((y_true - y_hat) / (np.abs(y_true) + np.abs(y_hat))))


@multivariate_support
def mase(actual_series: TimeSeries,
         pred_series: TimeSeries,
         insample: TimeSeries,
         m: Optional[int] = 1,
         intersect: bool = True,
         reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Mean Absolute Scaled Error (MASE).

    See `Mean absolute scaled error wikipedia page <https://en.wikipedia.org/wiki/Mean_absolute_scaled_error>`_
    for details about the MASE and how it is computed.

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    insample
        The training series used to forecast `pred_series` .
        This series serves to compute the scale of the error obtained by a naive forecaster on the training data.
    m
        Optionally, the seasonality to use for differencing.
        `m=1` corresponds to the non-seasonal MASE, whereas `m>1` corresponds to seasonal MASE.
        If `m=None`, it will be tentatively inferred
        from the auto-correlation function (ACF). It will fall back to a value of 1 if this fails.
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Raises
    ------
    ValueError
        If the `insample` series is periodic ( :math:`X_t = X_{t-m}` )

    Returns
    -------
    float
        The Mean Absolute Scaled Error (MASE)
    """

    raise_if_not(insample.end_time() + insample.freq() == pred_series.start_time(),
                 "The pred_series must be the forecast of the insample series", logger)
    if m is None:
        test_season, m = check_seasonality(insample)
        if not test_season:
            warn("No seasonality found when computing MASE. Fixing the period to 1.", UserWarning)
            m = 1
    y_true, y_hat = _get_values_or_raise(actual_series, pred_series, intersect)
    x_t = insample.values()
    errors = np.abs(y_true - y_hat)
    scale = np.mean(np.abs(x_t[m:] - x_t[:-m]))
    raise_if_not(not np.isclose(scale, 0), "cannot use MASE with periodical signals", logger)
    return np.mean(errors / scale)


@multivariate_support
def ope(actual_series: TimeSeries,
        pred_series: TimeSeries,
        intersect: bool = True,
        reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Overall Percentage Error (OPE).

    Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
    both of length :math:`T`, it is a percentage value computed as

    .. math:: 100 \\cdot \\left| \\frac{\\sum_{t=1}^{T}{y_t}
              - \\sum_{t=1}^{T}{\\hat{y}_t}}{\\sum_{t=1}^{T}{y_t}} \\right|.

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Raises
    ------
    ValueError
        If :math:`\\sum_{t=1}^{T}{y_t} = 0`.

    Returns
    -------
    float
        The Overall Percentage Error (OPE)
    """

    y_true, y_pred = _get_values_or_raise(actual_series, pred_series, intersect)
    y_true_sum, y_pred_sum = np.sum(y_true), np.sum(y_pred)
    raise_if_not(y_true_sum > 0, 'The series of actual value cannot sum to zero when computing OPE.', logger)
    return np.abs((y_true_sum - y_pred_sum) / y_true_sum) * 100.


@multivariate_support
def marre(actual_series: TimeSeries,
          pred_series: TimeSeries,
          intersect: bool = True,
          reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Mean Absolute Ranged Relative Error (MARRE).

    Given a time series of actual values :math:`y_t` and a time series of predicted values :math:`\\hat{y}_t`
    both of length :math:`T`, it is a percentage value computed as

    .. math:: 100 \\cdot \\frac{1}{T} \\sum_{t=1}^{T} {\\left| \\frac{y_t - \\hat{y}_t} {\\max_t{y_t} -
              \\min_t{y_t}} \\right|}

    Parameters
    ----------
    actual_series
        The series of actual values
    pred_series
        The series of predicted values
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Raises
    ------
    ValueError
        If :math:`\\max_t{y_t} = \\min_t{y_t}`.

    Returns
    -------
    float
        The Mean Absolute Ranged Relative Error (MARRE)
    """

    y_true, y_hat = _get_values_or_raise(actual_series, pred_series, intersect)
    raise_if_not(y_true.max() > y_true.min(), 'The difference between the max and min values must be strictly'
                 'positive to compute the MARRE.', logger)
    true_range = y_true.max() - y_true.min()
    return 100. * np.mean(np.abs((y_true - y_hat) / true_range))


@multivariate_support
def r2_score(series1: TimeSeries,
             series2: TimeSeries,
             intersect: bool = True,
             reduction: Callable[[np.ndarray], float] = np.mean) -> float:
    """ Coefficient of Determination :math:`R^2`.

    See `Coefficient of determination wikipedia page <https://en.wikipedia.org/wiki/Coefficient_of_determination>`_
    for details about the :math:`R^2` score and how it is computed.
    Please note that this metric is not symmetric, `series1` should correspond to the ground truth series,
    whereas `series2` should correspond to the predicted series.

    Parameters
    ----------
    series1
        The first time series. This should correspond to the ground truth values.
    series2
        The second time series. This should correspond to the predicted values.
    intersect
        For time series that are overlapping in time without having the same time index, setting `intersect=True`
        will consider the values only over their common time interval (intersection in time).
    reduction
        Function taking as input a np.ndarray and returning a scalar value. This function is used to aggregate
        the metrics of different components in case of multivariate TimeSeries instances.

    Returns
    -------
    float
        The Coefficient of Determination :math:`R^2`
    """

    y1, y2 = _get_values_or_raise(series1, series2, intersect)
    ss_errors = np.sum((y1 - y2) ** 2)
    y_hat = y1.mean()
    ss_tot = np.sum((y1 - y_hat) ** 2)
    return 1 - ss_errors / ss_tot



import unittest
import numpy as np
import pandas as pd
import random
import logging

from darts.metrics import mape, r2_score
from darts.utils.timeseries_generation import (
    linear_timeseries as lt,
    sine_timeseries as st,
    random_walk_timeseries as rt,
    constant_timeseries as ct,
    gaussian_timeseries as gt
)
from darts.models import (
    Theta,
    FFT,
    ExponentialSmoothing,
    NaiveSeasonal,
    StandardRegressionModel,
    NaiveDrift,
)

from ..logging import get_logger
logger = get_logger(__name__)

try:
    from ..models import TCNModel
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch models are not installed - will not be tested for backtesting')
    TORCH_AVAILABLE = False


def compare_best_against_random(model_class, params, series):

    # instantiate best model in expanding window mode
    best_model_1 = model_class.gridsearch(params, series, forecast_horizon=10, metric=mape)

    # instantiate best model in split mode
    train, val = series.split_before(series.time_index()[-10])
    best_model_2 = model_class.gridsearch(params, train, val_target_series=val, metric=mape)

    # intantiate model with random parameters from 'params'
    random_param_choice = {}
    for key in params.keys():
        random_param_choice[key] = random.choice(params[key])
    random_model = model_class(**random_param_choice)

    # perform backtest forecasting on both models
    best_forecast_1 = best_model_1.backtest(series, start=series.time_index()[-21], forecast_horizon=10)
    random_forecast_1 = random_model.backtest(series, start=series.time_index()[-21], forecast_horizon=10)

    # perform train/val evaluation on both models
    best_model_2.fit(train)
    best_forecast_2 = best_model_2.predict(len(val))
    random_model = model_class(**random_param_choice)
    random_model.fit(train)
    random_forecast_2 = random_model.predict(len(val))

    # check whether best models are at least as good as random models
    expanding_window_ok = mape(best_forecast_1, series) <= mape(random_forecast_1, series)
    split_ok = mape(best_forecast_2, val) <= mape(random_forecast_2, val)

    return expanding_window_ok and split_ok


class BacktestingTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        logging.disable(logging.CRITICAL)

    def test_backtest_forecasting(self):
        linear_series = lt(length=50)
        linear_series_multi = linear_series.stack(linear_series)

        # univariate model + univariate series
        pred = NaiveDrift().backtest(linear_series, None, pd.Timestamp('20000201'), 3)
        self.assertEqual(r2_score(pred, linear_series), 1.0)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=pd.Timestamp('20000217'), forecast_horizon=3)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=pd.Timestamp('20000217'), forecast_horizon=3,
                                  trim_to_series=True)
        NaiveDrift().backtest(linear_series, None, start=pd.Timestamp('20000216'), forecast_horizon=3)
        NaiveDrift().backtest(linear_series, None, pd.Timestamp('20000217'), forecast_horizon=3, trim_to_series=False)

        # Using forecast_horizon default value
        NaiveDrift().backtest(linear_series, None, start=pd.Timestamp('20000216'))
        NaiveDrift().backtest(linear_series, None, pd.Timestamp('20000217'), trim_to_series=False)

        # Using an int or float value for start
        NaiveDrift().backtest(linear_series, None, start=30)
        NaiveDrift().backtest(linear_series, None, start=0.7, trim_to_series=False)

        # Using invalid start and/or forecast_horizon values
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=0.7, forecast_horizon=-1)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, 0.7, -1)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=100)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=1.2)
        with self.assertRaises(TypeError):
            NaiveDrift().backtest(linear_series, None, start='wrong type')

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, None, start=49, forecast_horizon=2, trim_to_series=True)

        # univariate model + multivariate series
        with self.assertRaises(AssertionError):
            NaiveDrift().backtest(linear_series_multi, None, pd.Timestamp('20000201'), 3)

        # multivariate model + univariate series
        if TORCH_AVAILABLE:
            tcn_model = TCNModel(batch_size=1, n_epochs=1)
            pred = tcn_model.backtest(linear_series, None, pd.Timestamp('20000125'), 3, verbose=False)
            self.assertEqual(pred.width, 1)

            # multivariate model + multivariate series
            with self.assertRaises(ValueError):
                tcn_model.backtest(linear_series_multi, None, pd.Timestamp('20000125'), 3, verbose=False)
            tcn_model = TCNModel(batch_size=1, n_epochs=1, input_size=2, output_length=3)
            with self.assertRaises(ValueError):
                tcn_model.backtest(linear_series_multi, None, pd.Timestamp('20000125'), 3, verbose=False,
                                   use_full_output_length=False)
            pred = tcn_model.backtest(linear_series_multi, linear_series_multi[['0']], pd.Timestamp('20000125'), 1,
                                      verbose=False, use_full_output_length=True)
            self.assertEqual(pred.width, 1)
            pred = tcn_model.backtest(linear_series_multi, linear_series_multi[['1']], pd.Timestamp('20000125'), 3,
                                      verbose=False, use_full_output_length=True)
            self.assertEqual(pred.width, 1)
            tcn_model = TCNModel(batch_size=1, n_epochs=1, input_size=2, output_length=3, output_size=2)
            pred = tcn_model.backtest(linear_series_multi, linear_series_multi, pd.Timestamp('20000125'), 3,
                                      verbose=False, use_full_output_length=True)
            self.assertEqual(pred.width, 2)

    def test_backtest_regression(self):
        gaussian_series = gt(mean=2, length=50)
        sine_series = st(length=50)
        features = [gaussian_series + sine_series, gaussian_series]
        features_multivariate = [(gaussian_series + sine_series).stack(gaussian_series), gaussian_series]
        target = st(length=50)

        # univariate feature test
        pred = StandardRegressionModel(15).backtest(features, target, pd.Timestamp('20000201'), 3)
        self.assertEqual(r2_score(pred, target), 1.0)

        # Using an int or float value for start
        pred = StandardRegressionModel(15).backtest(features, target, start=30, forecast_horizon=3)
        self.assertEqual(r2_score(pred, target), 1.0)

        pred = StandardRegressionModel(15).backtest(features, target, start=0.5, forecast_horizon=3)
        self.assertEqual(r2_score(pred, target), 1.0)

        # Using a too small start value
        with self.assertRaises(ValueError):
            StandardRegressionModel(15).backtest(features, target, start=0, forecast_horizon=3)

        with self.assertRaises(ValueError):
            StandardRegressionModel(15).backtest(features, target, start=0.01, forecast_horizon=3)

        # Using StandardRegressionModel's start default value
        pred = StandardRegressionModel(15).backtest(features, target, forecast_horizon=3)
        self.assertEqual(r2_score(pred, target), 1.0)

        # multivariate feature test
        pred = StandardRegressionModel(15).backtest(features_multivariate, target, pd.Timestamp('20000201'), 3)
        self.assertEqual(r2_score(pred, target), 1.0)

        # multivariate target
        pred = StandardRegressionModel(15).backtest(features_multivariate, target.stack(target),
                                                    pd.Timestamp('20000201'), 3)
        self.assertEqual(r2_score(pred, target.stack(target)), 1.0)

        # multivariate target with stride
        pred = StandardRegressionModel(15).backtest(features_multivariate, target.stack(target),
                                                    pd.Timestamp('20000201'), 3, stride=3)
        self.assertEqual(r2_score(pred, target.stack(target)), 1.0)
        self.assertEqual((pred.time_index()[1] - pred.time_index()[0]).days, 3)

    def test_gridsearch(self):

        np.random.seed(1)
        ts_length = 50
        dummy_series = (
            lt(length=ts_length, end_value=10) + st(length=ts_length, value_y_offset=10) + rt(length=ts_length)
        )

        theta_params = {'theta': list(range(3, 10))}
        self.assertTrue(compare_best_against_random(Theta, theta_params, dummy_series))

        fft_params = {'nr_freqs_to_keep': [10, 50, 100], 'trend': [None, 'poly', 'exp']}
        self.assertTrue(compare_best_against_random(FFT, fft_params, dummy_series))

        es_params = {'seasonal_periods': list(range(5, 10))}
        self.assertTrue(compare_best_against_random(ExponentialSmoothing, es_params, dummy_series))

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_multi(self):
        dummy_series = st(length=40, value_y_offset=10).stack(lt(length=40, end_value=20))
        tcn_params = {
            'n_epochs': [1],
            'batch_size': [1],
            'input_size': [2],
            'output_length': [3],
            'output_size': [2],
            'kernel_size': [2, 3, 4]
        }
        TCNModel.gridsearch(tcn_params, dummy_series, forecast_horizon=3, metric=mape,
                            use_full_output_length=True)

    def test_forecasting_residuals(self):
        model = NaiveSeasonal(K=1)

        # test zero residuals
        constant_ts = ct(length=20)
        residuals = model.residuals(constant_ts)
        np.testing.assert_almost_equal(residuals.univariate_values(), np.zeros(len(residuals)))

        # test constant, positive residuals
        linear_ts = lt(length=20)
        residuals = model.residuals(linear_ts)
        np.testing.assert_almost_equal(np.diff(residuals.univariate_values()), np.zeros(len(residuals) - 1))
        np.testing.assert_array_less(np.zeros(len(residuals)), residuals.univariate_values())



"""
darts
-----
"""

from .timeseries import TimeSeries
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cycler

# Enums
from enum import Enum


class SeasonalityMode(Enum):
    MULTIPLICATIVE = 'multiplicative'
    ADDITIVE = 'additive'
    NONE = None


class TrendMode(Enum):
    LINEAR = 'linear'
    EXPONENTIAL = 'exponential'


class ModelMode(Enum):
    MULTIPLICATIVE = 'multiplicative'
    ADDITIVE = 'additive'


__version__ = '0.3.0'

colors = cycler(color=['black', '003DFD', 'b512b8', '11a9ba', '0d780f', 'f77f07', 'ba0f0f'])

u8plots_mplstyle = {
    'font.family' : 'Arial',
    'figure.facecolor' : '#f0f0f0',
    'axes.facecolor' : '#f0f0f0',
    'axes.prop_cycle' : colors,
    'lines.linewidth' : 1.3,
    'axes.spines.top' : False,
    'axes.spines.right' : False,
    'axes.spines.bottom' : False,
    'axes.spines.left' : False,
    'axes.linewidth' : 1,
    'axes.edgecolor' : 'black',
    'xtick.bottom' : False,
    'ytick.left' : False,
    'axes.grid' : True,
    'grid.color' : '#d0d0d0',
    'grid.alpha' : 0.5,
    'legend.frameon' : False}

mpl.rcParams.update(u8plots_mplstyle)



"""
Forecasting Model Base Class
----------------------------

A forecasting model captures the future values of a time series as a function of the past as follows:

.. math:: y_{t+1} = f(y_t, y_{t-1}, ..., y_1),

where :math:`y_t` represents the time series' value(s) at time :math:`t`.
"""

from typing import Optional, Tuple, Union, Any, Callable
from types import SimpleNamespace
from itertools import product
from abc import ABC, abstractmethod
import numpy as np
import pandas as pd

from ..timeseries import TimeSeries
from ..logging import get_logger, raise_log, raise_if_not
from ..utils import _build_tqdm_iterator, _with_sanity_checks, _get_timestamp_at_point, _backtest_general_checks
from .. import metrics

logger = get_logger(__name__)


class ForecastingModel(ABC):
    """ The base class for all forecasting models.

    All implementations of forecasting have to implement the `fit()` and `predict()` methods defined below.

    Attributes
    ----------
    training_series
        reference to the `TimeSeries` used for training the model.
    target_series
        reference to the `TimeSeries` used as target to train the model.
    """
    @abstractmethod
    def __init__(self):
        # Stores training date information:
        self.training_series: TimeSeries = None
        self.target_series: TimeSeries = None

        # state
        self._fit_called = False

    @abstractmethod
    def fit(self) -> None:
        """ Fits/trains the model on the provided series

        Implements behavior that should happen when calling the `fit` method of every forcasting model regardless of
        wether they are univariate or multivariate.
        """
        for series in (self.training_series, self.target_series):
            if series is not None:
                raise_if_not(len(series) >= self.min_train_series_length,
                             "Train series only contains {} elements but {} model requires at least {} entries"
                             .format(len(series), str(self), self.min_train_series_length))
        self._fit_called = True

    @abstractmethod
    def predict(self, n: int) -> TimeSeries:
        """ Predicts values for a certain number of time steps after the end of the training series

        Parameters
        ----------
        n
            The number of time steps after the end of the training time series for which to produce predictions

        Returns
        -------
        TimeSeries
            A time series containing the `n` next points, starting after the end of the training time series
        """

        if (not self._fit_called):
            raise_log(Exception('fit() must be called before predict()'), logger)

    @property
    def min_train_series_length(self) -> int:
        """
        Class property defining the minimum required length for the training series.
        This function/property should be overridden if a value higher than 3 is required.
        """
        return 3

    def _generate_new_dates(self, n: int) -> pd.DatetimeIndex:
        """
        Generates `n` new dates after the end of the training set
        """
        new_dates = [
            (self.training_series.time_index()[-1] + (i * self.training_series.freq())) for i in range(1, n + 1)
        ]
        return pd.DatetimeIndex(new_dates, freq=self.training_series.freq_str())

    def _build_forecast_series(self,
                               points_preds: np.ndarray) -> TimeSeries:
        """
        Builds a forecast time series starting after the end of the training time series, with the
        correct time index.
        """

        time_index = self._generate_new_dates(len(points_preds))

        return TimeSeries.from_times_and_values(time_index, points_preds, freq=self.training_series.freq())

    def _backtest_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Sanity checks for the backtest function

        Parameters
        ----------
        args
            The args parameter(s) provided to the backtest function.
        kwargs
            The kwargs paramter(s) provided to the backtest function.

        Raises
        ------
        ValueError
            when a check on the parameter does not pass.
        """
        # parse args and kwargs
        training_series = args[0]
        n = SimpleNamespace(**kwargs)

        # check target and training series
        if n.target_series is None:
            target_series = training_series
        else:
            target_series = n.target_series

        raise_if_not(all(training_series.time_index() == target_series.time_index()), "the target and training series"
                     " must have the same time indices.")

        _backtest_general_checks(training_series, kwargs)

    def _backtest_model_specific_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Method to be overriden in subclass for model specific sanity checks"""
        pass

    @_with_sanity_checks("_backtest_sanity_checks", "_backtest_model_specific_sanity_checks")
    def backtest(self,
                 training_series: TimeSeries,
                 target_series: Optional[TimeSeries] = None,
                 start: Union[pd.Timestamp, float, int] = 0.7,
                 forecast_horizon: int = 1,
                 stride: int = 1,
                 retrain: bool = True,
                 trim_to_series: bool = True,
                 verbose: bool = False,
                 use_full_output_length: Optional[bool] = None) -> TimeSeries:
        """ Retrain and forecast values pointwise with an expanding training window over `series`.

        To this end, it repeatedly builds a training set from the beginning of `series`. It trains the current model on
        the training set, emits a (point) prediction for a fixed forecast horizon, and then moves the end of the
        training set forward by `stride` time steps. The resulting predictions are then returned.

        Unless `retrain` is set to False, this always re-trains the models on the entire available history,
        corresponding an expending window strategy.

        If `retrain` is set to False (useful for models with many parameter such as `TorchForecastingModel` instances
        like `RNNModel` and `TCNModel`), the model will only be trained on the initial training window
        (up to `start` time stamp), and only if it has not been trained before. Then, at every iteration, the
        newly expanded input sequence will be fed to the model to produce the new output.

        Parameters
        ----------
        training_series
            The training time series on which to backtest
        target_series
            The target time series on which to backtest. This parameter is only relevant for
            `MultivariateForecastingModel` instances. It allows for training on one `training_series`
            and predicting another `target_series`. In many multivariate forecasting problems, the
            `target_series` would constitute a subset of the components of the `training_series`.
            However, any combination of univariate and multivariate series is allowed here, as long
            as the indices all match up.
        start
            The first prediction time, at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `training_series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        forecast_horizon
            The forecast horizon for the point prediction
        stride
            The number of time steps (the unit being the frequency of `series`) between two consecutive predictions.
        retrain
            Whether to retrain the model for every prediction or not. Currently only `TorchForecastingModel`
            instances such as `RNNModel` and `TCNModel` support setting `retrain` to `False`.
        use_full_output_length
            Optionally, if the model is an instance of `TorchForecastingModel`, this argument will be passed along
            as argument to the `predict` method of the model. Otherwise, if this value is set and the model is not an
            instance of `TorchForecastingModel`, this will cause an error.
        trim_to_series
            Whether the predicted series has the end trimmed to match the end of the main series
        verbose
            Whether to print progress

        Returns
        -------
        TimeSeries
            A time series containing the forecast values for `target_series`, when successively applying the specified
            model with the specified forecast horizon.
        """
        # handle case where target_series not specified
        if target_series is None:
            target_series = training_series

        # construct predict kwargs dictionary
        predict_kwargs = {}
        if use_full_output_length is not None:
            predict_kwargs['use_full_output_length'] = use_full_output_length

        # construct fit function (used to ignore target series for univariate models)
        if isinstance(self, MultivariateForecastingModel):
            fit_function = self.fit
        else:
            fit_function = lambda train, target, **kwargs: self.fit(train, **kwargs)  # noqa: E731

        # prepare the start parameter -> pd.Timestamp
        start = _get_timestamp_at_point(start, training_series)

        # build the prediction times in advance (to be able to use tqdm)
        if trim_to_series:
            last_pred_time = training_series.time_index()[-forecast_horizon - stride]
        else:
            last_pred_time = training_series.time_index()[-stride - 1]

        pred_times = [start]
        while pred_times[-1] <= last_pred_time:
            pred_times.append(pred_times[-1] + training_series.freq() * stride)

        # iterate and predict pointwise
        values = []
        times = []

        iterator = _build_tqdm_iterator(pred_times, verbose)

        if not retrain and not self._fit_called:
            fit_function(training_series.drop_after(start), target_series.drop_after(start), verbose=verbose)

        for pred_time in iterator:
            train = training_series.drop_after(pred_time)  # build the training series
            target = target_series.drop_after(pred_time)  # build the target series
            if (retrain):
                fit_function(train, target)
                pred = self.predict(forecast_horizon, **predict_kwargs)
            else:
                pred = self.predict(forecast_horizon, input_series=train, **predict_kwargs)
            values.append(pred.values()[-1])  # store the N-th point
            times.append(pred.end_time())  # store the N-th timestamp

        forecast = TimeSeries.from_times_and_values(pd.DatetimeIndex(times), np.array(values),
                                                    freq=training_series.freq() * stride)

        return forecast

    @classmethod
    def gridsearch(model_class,
                   parameters: dict,
                   training_series: TimeSeries,
                   target_series: Optional[TimeSeries] = None,
                   forecast_horizon: Optional[int] = None,
                   start: Union[pd.Timestamp, float, int] = 0.5,
                   use_full_output_length: Optional[bool] = None,
                   val_target_series: Optional[TimeSeries] = None,
                   use_fitted_values: bool = False,
                   metric: Callable[[TimeSeries, TimeSeries], float] = metrics.mape,
                   verbose=False) -> TimeSeries:
        """ A function for finding the best hyperparameters.

        This function has 3 modes of operation: Expanding window mode, split mode and fitted value mode.
        The three modes of operation evaluate every possible combination of hyperparameter values
        provided in the `parameters` dictionary by instantiating the `model_class` subclass
        of ForecastingModel with each combination, and returning the best-performing model with regards
        to the `metric` function. The `metric` function is expected to return an error value,
        thus the model resulting in the smallest `metric` output will be chosen.
        The relationship of the training data and test data depends on the mode of operation.

        Expanding window mode (activated when `forecast_horizon` is passed):
        For every hyperparameter combination, the model is repeatedly trained and evaluated on different
        splits of `training_series` and `target_series`. This process is accomplished by using
        `ForecastingModel.backtest` as a subroutine to produce historic forecasts starting from `start`
        that are compared against the ground truth values of `training_series` or `target_series`, if
        specifed.
        Note that the model is retrained for every single prediction, thus this mode is slower.

        Split window mode (activated when `val_series` is passed):
        This mode will be used when the `val_series` argument is passed.
        For every hyperparameter combination, the model is trained on `training_series` + `target_series` and
        evaluated on `val_series`.

        Fitted value mode (activated when `use_fitted_values` is set to `True`):
        For every hyperparameter combination, the model is trained on `training_series` + `target_series`
        and evaluated on the resulting fitted values.
        Not all models have fitted values, and this method raises an error if `model.fitted_values` does not exist.
        The fitted values are the result of the fit of the model on the training series. Comparing with the
        fitted values can be a quick way to assess the model, but one cannot see if the model overfits or underfits.


        Parameters
        ----------
        model_class
            The ForecastingModel subclass to be tuned for 'series'.
        parameters
            A dictionary containing as keys hyperparameter names, and as values lists of values for the
            respective hyperparameter.
        training_series
            The TimeSeries instance used as input for training.
        target_series
            The TimeSeries instance used as target for training (and also validation in expanding window mode).
        forecast_horizon
            The integer value of the forecasting horizon used in expanding window mode.
        start
            The `int`, `float` or `pandas.Timestamp` that represents the starting point in the time index
            of `training_series` from which predictions will be made to evaluate the model.
            For a detailed description of how the different data types are interpreted, please see the documentation
            for `ForecastingModel.backtest`.
        use_full_output_length
            This should only be set if `model_class` is equal to `TorchForecastingModel`.
            This argument will be passed along to the predict method of `TorchForecastingModel`.
        val_target_series
            The TimeSeries instance used for validation in split mode.
        use_fitted_values
            If `True`, uses the comparison with the fitted values.
            Raises an error if `fitted_values` is not an attribute of `model_class`.
        metric:
            A function that takes two TimeSeries instances as inputs and returns a float error value.
        verbose:
            Whether to print progress.

        Returns
        -------
        ForecastingModel
            An untrained 'model_class' instance with the best-performing hyperparameters from the given selection.
        """
        raise_if_not((forecast_horizon is not None) + (val_target_series is not None) + use_fitted_values == 1,
                     "Please pass exactly one of the arguments 'forecast_horizon', "
                     "'val_target_series' or 'use_fitted_values'.", logger)

        # check target and training series
        if target_series is None:
            target_series = training_series
        raise_if_not(all(training_series.time_index() == target_series.time_index()), "the target and training series"
                     " must have the same time indices.")

        # construct predict kwargs dictionary
        predict_kwargs = {}
        if use_full_output_length is not None:
            predict_kwargs['use_full_output_length'] = use_full_output_length

        if use_fitted_values:
            raise_if_not(hasattr(model_class(), "fitted_values"), "The model must have a fitted_values attribute"
                         " to compare with the train TimeSeries", logger)

        elif val_target_series is not None:
            raise_if_not(training_series.width == val_target_series.width, "Training and validation series require the"
                         " same number of components.", logger)

        min_error = float('inf')
        best_param_combination = {}

        # compute all hyperparameter combinations from selection
        params_cross_product = list(product(*parameters.values()))

        # iterate through all combinations of the provided parameters and choose the best one
        iterator = _build_tqdm_iterator(params_cross_product, verbose)
        for param_combination in iterator:
            param_combination_dict = dict(list(zip(parameters.keys(), param_combination)))
            model = model_class(**param_combination_dict)
            if use_fitted_values:  # fitted value mode
                model.fit(training_series)
                fitted_values = TimeSeries.from_times_and_values(training_series.time_index(), model.fitted_values)
                error = metric(fitted_values, target_series)
            elif val_target_series is None:  # expanding window mode
                backtest_forecast = model.backtest(training_series, target_series, start,
                                                   forecast_horizon, use_full_output_length=use_full_output_length)
                error = metric(backtest_forecast, target_series)
            else:  # split mode
                if isinstance(model, MultivariateForecastingModel):
                    model.fit(training_series, target_series)
                else:
                    model.fit(training_series)
                error = metric(model.predict(len(val_target_series), **predict_kwargs), val_target_series)
            if error < min_error:
                min_error = error
                best_param_combination = param_combination_dict
        logger.info('Chosen parameters: ' + str(best_param_combination))

        return model_class(**best_param_combination)

    def residuals(self,
                  series: TimeSeries,
                  forecast_horizon: int = 1,
                  verbose: bool = False) -> TimeSeries:
        """ A function for computing the residuals produced by the current model on a univariate time series.

        This function computes the difference between the actual observations from `series`
        and the fitted values vector p obtained by training the model on `series`.
        For every index i in `series`, p[i] is computed by training the model on
        series[:(i - `forecast_horizon`)] and forecasting `forecast_horizon` into the future.
        (p[i] will be set to the last value of the predicted vector.)
        The vector of residuals will be shorter than `series` due to the minimum
        training series length required by the model and the gap introduced by `forecast_horizon`.
        Note that the common usage of the term residuals implies a value for `forecast_horizon` of 1.

        Parameters
        ----------
        series
            The univariate TimeSeries instance which the residuals will be computed for.
        forecast_horizon
            The forecasting horizon used to predict each fitted value.
        verbose
            Whether to print progress.
        Returns
        -------
        TimeSeries
            The vector of residuals.
        """

        series._assert_univariate()

        # get first index not contained in the first training set
        first_index = series.time_index()[self.min_train_series_length]

        # compute fitted values
        p = self.backtest(series, None, first_index, forecast_horizon, 1, True, verbose=verbose)

        # compute residuals
        series_trimmed = series.slice_intersect(p)
        residuals = series_trimmed - p

        return residuals


class UnivariateForecastingModel(ForecastingModel):
    """The base class for univariate forecasting models."""
    @abstractmethod
    def fit(self, series: TimeSeries) -> None:
        """ Fits/trains the univariate model on selected univariate series.

        Implements behavior specific to calling the `fit` method on `UnivariateForecastingModel`.

        Parameters
        ----------
        series
            A **univariate** timeseries on which to fit the model.
        """
        series._assert_univariate()
        self.training_series = series
        self.target_series = series
        super().fit()


class MultivariateForecastingModel(ForecastingModel):
    """ The base class for multivariate forecasting models.
    """
    def _make_fitable_series(self,
                             training_series: TimeSeries,
                             target_series: Optional[TimeSeries] = None) -> Tuple[TimeSeries, TimeSeries]:
        """Perform checks and returns ready to be used training and target series"""
        if target_series is None:
            target_series = training_series

        # general checks on training / target series
        raise_if_not(all(training_series.time_index() == target_series.time_index()), "training and target "
                     "timeseries must have same time indices.")

        return training_series, target_series

    @abstractmethod
    def fit(self, training_series: TimeSeries, target_series: Optional[TimeSeries] = None) -> None:
        """ Fits/trains the multivariate model on the provided series with selected target components.

        Parameters
        ----------
        training_series
            The training time series on which to fit the model (can be multivariate or univariate).
        target_series
            The target values used as dependent variables when training the model
        """
        training_series, target_series = self._make_fitable_series(training_series, target_series)

        self.training_series = training_series
        self.target_series = target_series
        super().fit()



"""
ARIMA
-----

Models for ARIMA (Autoregressive integrated moving average) [1]_.
The implementations is wrapped around `statsmodels <https://github.com/statsmodels/statsmodels>`_.

References
----------
.. [1] https://wikipedia.org/wiki/Autoregressive_integrated_moving_average
"""

from statsmodels.tsa.arima_model import ARMA as staARMA
from statsmodels.tsa.arima_model import ARIMA as staARIMA

from .forecasting_model import UnivariateForecastingModel
from ..timeseries import TimeSeries
from ..logging import get_logger

logger = get_logger(__name__)


class ARIMA(UnivariateForecastingModel):
    def __init__(self, p: int = 12, d: int = 1, q: int = 0):
        """ ARIMA

        Parameters
        ----------
        p : int
            Order (number of time lags) of the autoregressive model (AR)
        d : int
            The order of differentiation; i.e., the number of times the data have had past values subtracted. (I)
        q : int
            The size of the moving average window (MA).
        """
        super().__init__()
        self.p = p
        self.d = d
        self.q = q
        self.model = None

    def __str__(self):
        return 'ARIMA({},{},{})'.format(self.p, self.d, self.q)

    def fit(self, series: TimeSeries):
        super().fit(series)
        series = self.training_series
        m = staARIMA(series.values(),
                     order=(self.p, self.d, self.q)) if self.d > 0 else staARMA(series.values(), order=(self.p, self.q))
        self.model = m.fit(disp=0)

    def predict(self, n):
        super().predict(n)
        forecast = self.model.forecast(steps=n)[0]
        return self._build_forecast_series(forecast)

    @property
    def min_train_series_length(self) -> int:
        return 30



"""
Additional util functions
-------------------------
"""
import pandas as pd

from ..timeseries import TimeSeries
from ..logging import raise_log, get_logger, raise_if_not, raise_if
from typing import List, Callable, TypeVar, Union
from IPython import get_ipython
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook
from functools import wraps
from types import SimpleNamespace
from inspect import signature, Parameter, getcallargs

logger = get_logger(__name__)


# TODO: we do not check the time index here
def retain_period_common_to_all(series: List[TimeSeries]) -> List[TimeSeries]:
    """
    Trims all series in the provided list, if necessary, so that the returned time series have
    a common span (corresponding to largest time sub-interval common to all series).

    Parameters
    ----------
    series
        The list of series to consider.

    Raises
    ------
    ValueError
        If no common time sub-interval exists

    Returns
    -------
    List[TimeSeries]
        A list of series, where each series have the same span
    """

    last_first = max(map(lambda s: s.start_time(), series))
    first_last = min(map(lambda s: s.end_time(), series))

    if last_first >= first_last:
        raise_log(ValueError('The provided time series must have nonzero overlap'), logger)

    return list(map(lambda s: s.slice(last_first, first_last), series))


def _build_tqdm_iterator(iterable, verbose):
    """
    Build an iterable, possibly using tqdm (either in notebook or regular mode)

    Parameters
    ----------
    iterable
    verbose

    Returns
    -------

    """

    def _isnotebook():
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    if verbose:
        if _isnotebook():
            iterator = tqdm_notebook(iterable)
        else:
            iterator = tqdm(iterable)
    else:
        iterator = iterable
    return iterator


# Types for sanity checks decorator
A = TypeVar('A')
B = TypeVar('B')
T = TypeVar('T')


def _with_sanity_checks(*sanity_check_methods: str) -> Callable[[Callable[[A, B], T]], Callable[[A, B], T]]:
    """
    Decorator allowing to specify some sanity check method(s) to be used on a class method.
    The decorator guarantees that args and kwargs from the method to sanitize will be available in the
    sanity check methods as specified in the sanitized method's signature, irrespective of how it was called.

    Parameters
    ----------
    *sanity_check_methods
        one or more sanity check methods that will be called with all the parameter of the decorated method.

    Returns
    -------
    A Callable corresponding to the decorated method.

    Examples
    --------
    class Model:
        def _a_sanity_check(self, *args, **kwargs):
            raise_if_not(kwargs['b'] == kwargs['c'], 'b must equal c', logger)

        @_with_sanity_checks("_a_sanity_check")
        def fit(self, a, b=0, c=0):
            # at this point we can safely assume that 'b' and 'c' are equal...
            ...
    """
    def decorator(method_to_sanitize: Callable[[A, B], T]) -> Callable[[A, B], T]:
        @wraps(method_to_sanitize)
        def sanitized_method(self, *args: A, **kwargs: B) -> T:
            for sanity_check_method in sanity_check_methods:
                # Convert all arguments into keyword arguments
                all_as_kwargs = getcallargs(method_to_sanitize, self, *args, **kwargs)

                # Then separate args from kwargs according to the function's signature
                only_args = all_as_kwargs.copy()
                only_kwargs = all_as_kwargs.copy()

                for param_name, param in signature(method_to_sanitize).parameters.items():
                    if param.default == Parameter.empty and param.kind != Parameter.VAR_POSITIONAL:
                        only_kwargs.pop(param_name)
                    else:
                        only_args.pop(param_name)

                only_args.pop('self')

                getattr(self, sanity_check_method)(*only_args.values(), **only_kwargs)
            return method_to_sanitize(self, *only_args.values(), **only_kwargs)
        return sanitized_method
    return decorator


def _backtest_general_checks(series, kwargs):
    """
    Performs checks common to ForecastingModel and RegressionModel backtest() methods

    Parameters
    ----------
    series
        Either training_series when called from ForecastingModel, or target_series if called from RegressionModel
    signature_params
        A dictionary of the signature parameters of the calling method, to get the default values
        Typically would be signature(self.backtest).parameters
    kwargs
        Params specified by the caller of backtest(), they take precedence over the arguments' default values
    """

    # parse kwargs
    n = SimpleNamespace(**kwargs)

    # check forecast horizon
    forecast_horizon = n.forecast_horizon

    raise_if_not(forecast_horizon > 0, 'The provided forecasting horizon must be a positive integer.', logger)

    # check start parameter
    if hasattr(n, 'start'):
        if isinstance(n.start, float):
            raise_if_not(n.start >= 0.0 and n.start < 1.0, '`start` should be between 0.0 and 1.0.', logger)
        elif isinstance(n.start, pd.Timestamp):
            raise_if(n.start not in series, '`start` timestamp must be an entry in the time series\' time index')
            raise_if(n.start == series.end_time(), '`start` timestamp is the last timestamp of the series', logger)
        elif isinstance(n.start, int):
            raise_if_not(n.start >= 0, logger)
            raise_if(n.start > len(series), '`start` index should be smaller than length of the series', logger)
        else:
            raise_log(TypeError("`start` needs to be either `float`, `int` or `pd.Timestamp`"), logger)

    start = _get_timestamp_at_point(n.start, series)

    raise_if(start == series.start_time(), '`start` corresponds to the first timestamp of the series, '
             'resulting in empty training set')

    # check that trim_to_series and start together form a valid combination
    trim_to_series = n.trim_to_series

    if trim_to_series:
        raise_if_not(start + series.freq() * forecast_horizon in series,
                     '`start` timestamp is too late in the series to make any predictions with'
                     '`trim_to_series` set to `True`.', logger)


def _get_timestamp_at_point(point: Union[pd.Timestamp, float, int], series: TimeSeries) -> pd.Timestamp:
    """
    Converts a point into a pandas.Timestamp in the time series

    Parameters
    ----------
    point
        This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
        In the case of `float`, the parameter will be treated as the proportion of the time series
        that should lie before the point.
        In the case of `int`, the parameter will be treated as an integer index to the time index of
        `series`.
        In case of `pandas.Timestamp`, `point` will be returned as is
    series
        The time series to index in
    """
    if isinstance(point, float):
        point_index = int((len(series.time_index()) - 1) * point)
        timestamp = series.time_index()[point_index]
    elif isinstance(point, int):
        timestamp = series[point].start_time()
    else:
        timestamp = point
    return timestamp



"""
Torch Forecasting Model
-----------------------
"""

import numpy as np
import os
import re
import math
from glob import glob
import shutil
from typing import Optional, Dict
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from torch.utils.tensorboard import SummaryWriter

from ..timeseries import TimeSeries
from ..utils import _build_tqdm_iterator
from ..utils.torch import random_method
from ..logging import raise_if_not, get_logger, raise_log, raise_if
from .forecasting_model import MultivariateForecastingModel

CHECKPOINTS_FOLDER = os.path.join('.darts', 'checkpoints')
RUNS_FOLDER = os.path.join('.darts', 'runs')

logger = get_logger(__name__)


def _get_checkpoint_folder(work_dir, model_name):
    return os.path.join(work_dir, CHECKPOINTS_FOLDER, model_name)


def _get_runs_folder(work_dir, model_name):
    return os.path.join(work_dir, RUNS_FOLDER, model_name)


class _TimeSeriesSequentialDataset(Dataset):

    def __init__(self,
                 training_series: TimeSeries,
                 target_series: TimeSeries,
                 data_length: int = 1,
                 target_length: int = 1):
        """
        A PyTorch Dataset from a univariate TimeSeries.
        The Dataset iterates a moving window over the time series. The resulting slices contain `(data, target)`,
        where `data` is a 1-D sub-sequence of length `data_length` and target is the 1-D sub-sequence of length
        `target_length` following it in the time series.

        Parameters
        ----------
        training_series
            The time series to be included in the dataset.
        target_series
            The time series used as target.
        data_length
            The length of the training sub-sequences.
        target_length
            The length of the target sub-sequences, starting at the end of the training sub-sequence.
        """

        self.training_series_values = training_series.values()
        self.target_series_values = target_series.values()

        # self.series = torch.from_numpy(self.series).float()  # not possible to cast in advance
        self.len_series = len(training_series)
        self.data_length = len(training_series) - 1 if data_length is None else data_length
        self.target_length = target_length

        raise_if_not(self.data_length > 0,
                     "The input sequence length must be positive. It is {}".format(self.data_length),
                     logger)
        raise_if_not(self.target_length > 0,
                     "The output sequence length must be positive. It is {}".format(self.target_length),
                     logger)

    def __len__(self):
        return self.len_series - self.data_length - self.target_length + 1

    def __getitem__(self, index):
        # TODO: Cast to PyTorch tensors on the right device in advance
        idx = index % (self.len_series - self.data_length - self.target_length + 1)
        data = self.training_series_values[idx:idx + self.data_length]
        target = self.target_series_values[idx + self.data_length:idx + self.data_length + self.target_length]
        return torch.from_numpy(data).float(), torch.from_numpy(target).float()


class _TimeSeriesShiftedDataset(Dataset):

    def __init__(self,
                 training_series: TimeSeries,
                 target_series: TimeSeries,
                 length: int = 3,
                 shift: int = 1):
        """
        A PyTorch Dataset from a univariate TimeSeries.
        The Dataset iterates a moving window over the time series. The resulting slices contain `(data, target)`,
        where `data` and `target` are both 1-D sub-sequences of length `ength`. The sequence contained in
        target is shifted forward by `shift` positions, meaning that `target` contains the last
        `length` - `shift` entries of `data` and then the `shift` following ones.

        Parameters
        ----------
        training_series
            The time series to be included in the dataset.
        target_series
            The time series used as target.
        length
            The length of the training and target sub-sequences.
        shift
            The number of positions that the target sequence is shifted forward compared to the training sequence.
        """

        self.training_series_values = training_series.values()
        self.target_series_values = target_series.values()
        self.len_series = len(training_series)
        self.length = len(training_series) - 1 if length is None else length
        self.shift = shift

        raise_if_not(self.length > 0,
                     "The input sequence length must be positive. It is {}".format(self.length),
                     logger)

        raise_if_not(self.shift > 0,
                     "The shift value must be positive. It is {}".format(self.length),
                     logger)

    def __len__(self):
        return self.len_series - self.length - self.shift + 1

    def __getitem__(self, index):
        idx = index % self.__len__()
        data = self.training_series_values[idx:idx + self.length]
        target = self.target_series_values[idx + self.shift:idx + self.length + self.shift]
        return torch.from_numpy(data).float(), torch.from_numpy(target).float()


class TorchForecastingModel(MultivariateForecastingModel):
    # TODO: add is_stochastic & reset methods
    # TODO: transparent support for multivariate time series
    def __init__(self,
                 batch_size: int = 32,
                 output_length: int = 1,
                 input_length: int = 10,
                 input_size: int = 1,
                 output_size: int = 1,
                 n_epochs: int = 800,
                 optimizer_cls: torch.optim.Optimizer = torch.optim.Adam,
                 optimizer_kwargs: Dict = None,
                 lr_scheduler_cls: torch.optim.lr_scheduler._LRScheduler = None,
                 lr_scheduler_kwargs: Optional[Dict] = None,
                 loss_fn: nn.modules.loss._Loss = nn.MSELoss(),
                 model_name: str = "torch_model_run",  # TODO uid
                 work_dir: str = os.getcwd(),
                 log_tensorboard: bool = False,
                 nr_epochs_val_period: int = 10,
                 torch_device_str: Optional[str] = None):

        """ Pytorch-based Forecasting Model.

        This class is meant to be inherited to create a new pytorch-based forecasting module.
        When subclassing this class, please make sure to set the self.model attribute
        in the __init__ function and then call super().__init__ while passing the kwargs.

        Parameters
        ----------
        output_length
            Number of time steps to be output by the forecasting module.
        input_length
            Number of past time steps that are fed to the forecasting module.
        input_size
            The dimensionality of the TimeSeries instances that will be fed to the fit function.
        output_size
            The dimensionality of the output time series.
        batch_size
            Number of time series (input and output sequences) used in each training pass.
        n_epochs
            Number of epochs over which to train the model.
        optimizer_cls
            The PyTorch optimizer class to be used (default: `torch.optim.Adam`).
        optimizer_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer (e.g., `{'lr': 1e-3}`)
            for specifying a learning rate. Otherwise the default values of the selected `optimizer_cls`
            will be used.
        lr_scheduler_cls
            Optionally, the PyTorch learning rate scheduler class to be used. Specifying `None` corresponds
            to using a constant learning rate.
        lr_scheduler_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer.
        loss_fn
            PyTorch loss function used for training (default: `torch.nn.MSELoss()`).
        model_name
            Name of the model. Used for creating the checkpoints and saving tensorboard data.
        work_dir
            Path of the working directory, where to save checkpoints and Tensorboard summaries.
            (default: current working directory).
        log_tensorboard
            If set, use Tensorboard to log the different parameters. The logs will be located in:
            `[work_dir]/.darts/runs/`.
        nr_epochs_val_period
            Number of epochs to wait before evaluating the validation loss (if a validation
            `TimeSeries` is passed to the `fit()` method).
        torch_device_str
            Optionally, a string indicating the torch device to use. (default: "cuda:0" if a GPU
            is available, otherwise "cpu")
        """

        super().__init__()

        raise_if_not(isinstance(self.model, nn.Module), 'Please make sure that self.model is set to a valid '
                     'nn.Module subclass when subclassing TorchForecastingModel',
                     logger)

        if torch_device_str is None:
            self.device = self._get_best_torch_device()
        else:
            self.device = torch.device(torch_device_str)

        self.input_length = input_length
        self.input_size = input_size
        self.output_length = output_length
        self.output_size = output_size
        self.log_tensorboard = log_tensorboard
        self.nr_epochs_val_period = nr_epochs_val_period

        self.model = self.model.to(self.device)
        self.model_name = model_name
        self.work_dir = work_dir

        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.from_scratch = True  # do we train the model from scratch

        # Define the loss function
        self.criterion = loss_fn

        # The tensorboard writer
        self.tb_writer = None

        # A utility function to create optimizer and lr scheduler from desired classes
        def _create_from_cls_and_kwargs(cls, kws):
            try:
                instance = cls(**kws)
            except (TypeError, ValueError) as e:
                raise_log(ValueError('Error when building the optimizer or learning rate scheduler;'
                                     'please check the provided class and arguments'
                                     '\nclass: {}'
                                     '\narguments (kwargs): {}'
                                     '\nerror:\n{}'.format(cls, kws, e)),
                          logger)
            return instance

        # Create the optimizer and (optionally) the learning rate scheduler
        if optimizer_kwargs is None:
            optimizer_kwargs = {'lr': 1e-3}  # we do not set this as a default argument to avoid mutable values
        optimizer_kwargs['params'] = self.model.parameters()
        self.optimizer = _create_from_cls_and_kwargs(optimizer_cls, optimizer_kwargs)

        if lr_scheduler_cls is not None:
            lr_scheduler_kwargs['optimizer'] = self.optimizer
            self.lr_scheduler = _create_from_cls_and_kwargs(lr_scheduler_cls, lr_scheduler_kwargs)
        else:
            self.lr_scheduler = None  # We won't use a LR scheduler

    @random_method
    def fit(self,
            training_series: TimeSeries,
            target_series: Optional[TimeSeries] = None,
            val_training_series: Optional[TimeSeries] = None,
            val_target_series: Optional[TimeSeries] = None,
            verbose: bool = False) -> None:
        """ Fit method for torch modules

        Parameters
        ----------
        training_series
            A series of training values used to predict the target series (can also include the target).
        target_series
            A series of target (dependent) value(s) predicted using the training_series if None specified, use the
            training series as target.
        val_training_series
            Optionally, a validation training time series, which will be used to compute the validation
            loss throughout training and keep track of the best performing models.
        val_target_series
            Optionally, a validation target time series, which will be used to compute the validation
            loss throughout training and keep track of the best performing models.
        verbose
            Optionally, whether to print progress.
        """
        super().fit(training_series, target_series)

        raise_if_not(self.training_series.width == self.input_size, "The number of components of the training series "
                     "must be equal to the `input_size` defined when instantiating the current model.", logger)
        raise_if_not(self.target_series.width == self.output_size, "The number of components in the target series must "
                     "be equal to the `output_size` defined when instantiating the current model.", logger)

        # perform checks on the validation series
        raise_if(val_training_series is None and val_target_series is not None, "`val_target_series` can not be "
                 "specified without a `val_training_series`.")
        if val_training_series is not None:
            val_training_series, val_target_series = self._make_fitable_series(val_training_series, val_target_series)
            raise_if_not(self.training_series.width == val_training_series.width, "training_series must have the "
                         "same number of component(s) as val_training_series.", logger)
            raise_if_not(self.target_series.width == val_target_series.width, "target_series must have the same number "
                         "of component(s) as val_target_series.", logger)

        if self.from_scratch:
            shutil.rmtree(_get_checkpoint_folder(self.work_dir, self.model_name), ignore_errors=True)

        # Prepare training data
        dataset = self._create_dataset(self.training_series, self.target_series)
        train_loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True, num_workers=0, pin_memory=True,
                                  drop_last=True)
        raise_if_not(len(train_loader) > 0,
                     'The provided training time series is too short for obtaining even one training point.',
                     logger)

        # Prepare validation data
        val_loader = None if val_training_series is None else self._prepare_validation_data(val_training_series,
                                                                                            val_target_series)

        # Prepare tensorboard writer
        tb_writer = self._prepare_tensorboard_writer()

        # Train model
        self._train(train_loader, val_loader, tb_writer, verbose)

        # Close tensorboard writer
        if tb_writer is not None:
            tb_writer.flush()
            tb_writer.close()

    def predict(self, n: int,
                use_full_output_length: bool = False,
                input_series: Optional[TimeSeries] = None) -> TimeSeries:

        """ Predicts values for a certain number of time steps after the end of the training series

        In the case of univariate training series, `n` can assume any integer value greater than 0.
        If `use_full_output_length` is set to `False`, the model will perform `n` predictions, where in each iteration
        the first predicted value is kept as output while at the same time being fed into the input for
        the next prediction (the first value of the previous input is discarded). This way, the input sequence
        'rolls over' by 1 step for every prediction in 'n'.
        If `use_full_output_length` is set to `True`, the model will predict not one, but `self.output_length` values
        in every iteration. This means that `ceil(n / self.output_length)` iterations will be required. After
        every iteration the input sequence 'rolls over' by `self.output_length` steps, meaning that the last
        `self.output_length` entries in the input sequence will correspond to the prediction of the previous
        iteration.

        In the case of multivariate training series, `n` cannot exceed `self.output_length` and `use_full_output_length`
        has to be set to `True`. In this case, only one iteration of predictions will be performed.

        Parameters
        ----------
        n
            The number of time steps after the end of the training time series for which to produce predictions
        use_full_output_length
            Boolean value indicating whether or not the full output sequence of the model prediction should be
            used to produce the output of this function.
        input_series
            Optionally, the input TimeSeries instance fed to the trained TorchForecastingModel to produce the
            prediction. If it is not passed, the training TimeSeries instance will be used as input.

        Returns
        -------
        TimeSeries
            A time series containing the `n` next points, starting after the end of the training time series
        """
        super().predict(n)

        raise_if_not(input_series is None or input_series.width == self.training_series.width,
                     "'input_series' must have same width as series used to fit model.", logger)

        raise_if_not(use_full_output_length or (self.training_series == self.target_series
                                                and self.training_series.width == 1),
                     "Please set 'use_full_output_length' to 'True' and 'n' smaller or equal to 'output_length'"
                     " when using a multivariate TimeSeries instance as input.", logger)

        # create input sequence for prediction
        pred_in = self._create_predict_input(input_series)

        # produce output
        self.model.eval()
        if (use_full_output_length):
            test_out = self._produce_predict_output_with_full_output_length(pred_in, n)
        else:
            test_out = self._produce_predict_output_with_single_output_steps(pred_in, n)

        test_out = np.stack(test_out)
        return self._build_forecast_series(test_out.reshape(n, -1))

    @property
    def first_prediction_index(self) -> int:
        """
        Returns the index of the first predicted within the output of self.model.
        """
        return 0

    def _create_dataset(self, training_series, target_series):
        return _TimeSeriesSequentialDataset(training_series, target_series, self.input_length, self.output_length)

    def _train(self,
               train_loader: DataLoader,
               val_loader: Optional[DataLoader],
               tb_writer: Optional[SummaryWriter],
               verbose: bool) -> None:
        """
        Performs the actual training
        :param train_loader: the training data loader feeding the training data and targets
        :param val_loader: optionally, a validation set loader
        :param tb_writer: optionally, a TensorBoard writer
        """

        best_loss = np.inf

        iterator = _build_tqdm_iterator(range(self.n_epochs), verbose)
        for epoch in iterator:
            epoch = epoch
            total_loss = 0
            total_loss_diff = 0

            for batch_idx, (data, target) in enumerate(train_loader):
                self.model.train()
                data, target = data.to(self.device), target.to(self.device)  # TODO: needed if done in dataset?
                output = self.model(data)
                loss = self.criterion(output, target)
                if self.output_length == 1:
                    loss_of_diff = self.criterion(output[1:] - output[:-1], target[1:] - target[:-1])
                else:
                    loss_of_diff = self.criterion(output[:, 1:] - output[:, :-1], target[:, 1:] - target[:, :-1])
                loss = loss + loss_of_diff
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                total_loss += loss.item()
                total_loss_diff += loss_of_diff.item()
            if self.lr_scheduler is not None:
                self.lr_scheduler.step()

            if tb_writer is not None:
                for name, param in self.model.named_parameters():
                    tb_writer.add_histogram(name + '/gradients', param.grad.data.cpu().numpy(), epoch)
                tb_writer.add_scalar("training/loss", total_loss / (batch_idx + 1), epoch)
                tb_writer.add_scalar("training/loss_diff", total_loss_diff / (batch_idx + 1), epoch)
                tb_writer.add_scalar("training/loss_total", (total_loss + total_loss_diff) / (batch_idx + 1), epoch)
                tb_writer.add_scalar("training/learning_rate", self._get_learning_rate(), epoch)

            self._save_model(False, _get_checkpoint_folder(self.work_dir, self.model_name), epoch)

            if epoch % self.nr_epochs_val_period == 0:
                training_loss = (total_loss + total_loss_diff) / (batch_idx + 1)  # TODO: do not use batch_idx
                if val_loader is not None:
                    validation_loss = self._evaluate_validation_loss(val_loader)
                    if tb_writer is not None:
                        tb_writer.add_scalar("validation/loss_total", validation_loss, epoch)

                    if validation_loss < best_loss:
                        best_loss = validation_loss
                        self._save_model(True, _get_checkpoint_folder(self.work_dir, self.model_name), epoch)

                    if verbose:
                        print("Training loss: {:.4f}, validation loss: {:.4f}".
                              format(training_loss, validation_loss), end="\r")
                elif verbose:
                    print("Training loss: {:.4f}".format(training_loss), end="\r")

    def _evaluate_validation_loss(self, val_loader: DataLoader):
        total_loss = 0
        total_loss_of_diff = 0
        self.model.eval()
        with torch.no_grad():
            for batch_idx, (data, target) in enumerate(val_loader):
                data, target = data.to(self.device), target.to(self.device)  # TODO: needed?
                output = self.model(data)
                loss = self.criterion(output, target)
                if self.output_length == 1:
                    loss_of_diff = self.criterion(output[1:] - output[:-1], target[1:] - target[:-1])
                else:
                    loss_of_diff = self.criterion(output[:, 1:] - output[:, :-1], target[:, 1:] - target[:, :-1])
                total_loss += loss.item()
                total_loss_of_diff += loss_of_diff.item()

        validation_loss = (total_loss + total_loss_of_diff) / (batch_idx + 1)
        return validation_loss

    def _save_model(self,
                    is_best: bool,
                    folder: str,
                    epoch: int):
        """
        Saves the whole torch model object to a file

        :param is_best: whether the model we're currently saving is the best (on validation set)
        :param folder:
        :param epoch:
        :return:
        """

        checklist = glob(os.path.join(folder, "checkpoint_*"))
        checklist = sorted(checklist, key=lambda x: float(re.findall(r'(\d+)', x)[-1]))
        filename = 'checkpoint_{0}.pth.tar'.format(epoch)
        os.makedirs(folder, exist_ok=True)
        filename = os.path.join(folder, filename)

        with open(filename, 'wb') as f:
            torch.save(self, f)

        if len(checklist) >= 5:
            # remove older files
            for chkpt in checklist[:-4]:
                os.remove(chkpt)
        if is_best:
            best_name = os.path.join(folder, 'model_best_{0}.pth.tar'.format(epoch))
            shutil.copyfile(filename, best_name)
            checklist = glob(os.path.join(folder, "model_best_*"))
            checklist = sorted(checklist, key=lambda x: float(re.findall(r'(\d+)', x)[-1]))
            if len(checklist) >= 2:
                # remove older files
                for chkpt in checklist[:-1]:
                    os.remove(chkpt)

    def _prepare_validation_data(self, val_training_series, val_target_series):
        val_dataset = self._create_dataset(val_training_series, val_target_series)
        val_loader = DataLoader(val_dataset, batch_size=self.batch_size, shuffle=False,
                                num_workers=0, pin_memory=True, drop_last=False)
        raise_if_not(len(val_dataset) > 0 and len(val_loader) > 0,
                     'The provided validation time series is too short for this model output length.',
                     logger)
        return val_loader

    def _prepare_tensorboard_writer(self):
        runs_folder = _get_runs_folder(self.work_dir, self.model_name)
        if self.log_tensorboard:
            if self.from_scratch:
                shutil.rmtree(runs_folder, ignore_errors=True)
                tb_writer = SummaryWriter(runs_folder)
                dummy_input = torch.empty(self.batch_size, self.input_length, self.input_size).to(self.device)
                tb_writer.add_graph(self.model, dummy_input)
            else:
                tb_writer = SummaryWriter(runs_folder, purge_step=self.start_epoch)
        else:
            tb_writer = None
        return tb_writer

    def _create_predict_input(self, input_series):
        if input_series is None:
            input_sequence = self.training_series.values()[-self.input_length:]
        else:
            raise_if_not(len(input_series) >= self.input_length,
                         "'input_series' must at least be as long as 'self.input_length'", logger)
            input_sequence = input_series.values()[-self.input_length:]
            super().fit(input_series)
        pred_in = torch.from_numpy(input_sequence).float().view(1, self.input_length, -1).to(self.device)

        return pred_in

    def _produce_predict_output_with_full_output_length(self, pred_in, n):
        test_out = []
        num_iterations = int(math.ceil(n / self.output_length))
        for i in range(num_iterations):
            raise_if_not(self.training_series.width == 1 or num_iterations == 1, "Only univariate time series"
                         " support predictions with n > output_length", logger)
            out = self.model(pred_in)
            if (num_iterations > 1):
                pred_in = pred_in.roll(-self.output_length, 1)
                pred_in[:, -self.output_length:, 0] = out[:, -self.output_length:].view(1, -1)
            test_out.append(out.cpu().detach().numpy()[:, -self.output_length:])
        test_out = np.concatenate(test_out, axis=1)
        test_out = test_out[:, :n, :]
        return test_out

    def _produce_predict_output_with_single_output_steps(self, pred_in, n):
        test_out = []
        for i in range(n):
            out = self.model(pred_in)
            pred_in = pred_in.roll(-1, 1)
            pred_in[:, -1, 0] = out[:, self.first_prediction_index]
            test_out.append(out.cpu().detach().numpy()[0, self.first_prediction_index])
        return test_out

    @staticmethod
    def load_from_checkpoint(model_name: str,
                             work_dir: str = os.getcwd(),
                             filename: str = None,
                             best: bool = True) -> 'TorchForecastingModel':
        """
        Load the model from the given checkpoint.
        if file is not given, will try to restore the most recent checkpoint.

        Parameters
        ----------
        model_name
            The name of the model (used to retrieve the checkpoints folder's name).
        work_dir
            Working directory (containing the checkpoints folder). Defaults to current working directory.
        filename
            The name of the checkpoint file. If not specified, use the most recent one.
        best
            If set, will retrieve the best model (according to validation loss) instead of the most recent one.

        Returns
        -------
        TorchForecastingModel
            The corresponding trained `TorchForecastingModel`.
        """

        checkpoint_dir = _get_checkpoint_folder(work_dir, model_name)

        # if filename is none, find most recent file in savepath that is a checkpoint
        if filename is None:
            path = os.path.join(checkpoint_dir, "model_best_*" if best else "checkpoint_*")
            checklist = glob(path)
            if len(checklist) == 0:
                raise_log(FileNotFoundError('There is no file matching prefix {} in {}'.format(
                          "model_best_*" if best else "checkpoint_*", checkpoint_dir)),
                          logger)
            filename = max(checklist, key=os.path.getctime)  # latest file TODO: check case where no files match
            filename = os.path.basename(filename)

        full_fname = os.path.join(checkpoint_dir, filename)
        print('loading {}'.format(filename))
        with open(full_fname, 'rb') as f:
            model = torch.load(f)
        return model

    def _get_best_torch_device(self):
        is_cuda = torch.cuda.is_available()
        if is_cuda:
            return torch.device("cuda:0")
        else:
            return torch.device("cpu")

    def _get_learning_rate(self):
        for p in self.optimizer.param_groups:
            return p['lr']



import unittest
import pandas as pd

from ..timeseries import TimeSeries
from ..logging import get_logger

logger = get_logger(__name__)

try:
    import torch
    from ..models.torch_forecasting_model import _TimeSeriesSequentialDataset
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch is not available. Torch Dataset will not be tested')
    TORCH_AVAILABLE = False

if TORCH_AVAILABLE:
    class DatasetTestCase(unittest.TestCase):
        __test__ = True
        times = pd.date_range('20130101', '20130410')
        pd_series = pd.Series(range(100), index=times)
        series: TimeSeries = TimeSeries.from_series(pd_series)

        def test_creation(self):
            # Cannot have train window <= 0
            with self.assertRaises(ValueError):
                _TimeSeriesSequentialDataset(self.series, self.series, -1)
            # Cannot have label window <= 0
            with self.assertRaises(ValueError):
                _TimeSeriesSequentialDataset(self.series, self.series, 1, -1)

        def test_content(self):
            # Can have no transformation
            dataset = _TimeSeriesSequentialDataset(self.series, self.series, 12, 4)
            self.assertEqual(len(dataset), 100 - 12 - 4 + 1)
            # correct types
            self.assertEqual(type(dataset[0][0]), torch.Tensor)
            self.assertEqual(dataset[0][0].dtype, torch.float32)
            # take correct elements
            self.assertEqual(dataset[5][0].numpy()[0], self.pd_series.values[5])
            self.assertEqual(dataset[5][1].numpy()[0], self.pd_series.values[5 + 12])
            # content same size from beginning to end
            self.assertEqual(dataset[0][1].size(), dataset[len(dataset) - 1][1].size())



"""
Regression Model Base Class
---------------------------

A regression model predicts values for a time series :math:`Y_t` as a function
of :math:`N` "features" time series :math:`X^i_t`:

.. math:: Y_t = f(X^1_t, ..., X^N_t),

where :math:`t` denotes the time step. Here, the function :math:`f()` is not necessarily linear.
"""

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import List, Iterable, Union, Any

from ..timeseries import TimeSeries
from ..logging import raise_if_not, get_logger, raise_log
from ..utils import _build_tqdm_iterator, _with_sanity_checks, _get_timestamp_at_point, _backtest_general_checks

logger = get_logger(__name__)


# TODO: Extend this to a "DynamicRegressiveModel" class, which acts on List[List[TimeSeries]].
# TODO: The first List[] would contain time-sliding lists of time series, letting the model
# TODO: be able to learn how to change weights over time. When len() of outer List[] is 0 it's a particular case
class RegressionModel(ABC):
    @abstractmethod
    def __init__(self):
        """ Regression Model.

            This is the base class for all regression models.
        """

        # Stores training date information:
        self.train_features: List[TimeSeries] = None
        self.train_target: TimeSeries = None

        # state
        self._fit_called = False

    @abstractmethod
    def fit(self, train_features: List[TimeSeries], train_target: TimeSeries) -> None:
        """ Fits/trains the model using the provided list of features time series and the target time series.

        Parameters
        ----------
        train_features
            A list of features time series, all of the same length as the target series
        train_target
            A target time series, of the same length as the features series
        """

        raise_if_not(len(train_features) > 0, 'Need at least one feature series', logger)
        raise_if_not(all([s.has_same_time_as(train_target) for s in train_features]),
                     'All provided time series must have the same time index', logger)
        self.train_features = train_features
        self.train_target = train_target
        self._fit_called = True

    @abstractmethod
    def predict(self, features: List[TimeSeries]) -> TimeSeries:
        """ Predicts values of the target time series, given a list of features time series

        Parameters
        ----------
        features
            The list of features time series, of the same length

        Returns
        -------
        TimeSeries
            A series containing the predicted targets, of the same length as the features series
        """

        if (not self._fit_called):
            raise_log(Exception('fit() must be called before predict()'), logger)

        length_ok = len(features) == len(self.train_features)
        dimensions_ok = all(features[i].width == self.train_features[i].width for i in range(len(features)))
        raise_if_not(length_ok and dimensions_ok,
                     'The number and dimensionalities of all given features must correspond to those used for'
                     ' training.', logger)

    def _backtest_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Sanity checks for the backtest function

        Parameters
        ----------
        args
            The args parameter(s) provided to the backtest function.
        kwargs
            The kwargs paramter(s) provided to the backtest function.

        Raises
        ------
        ValueError
            when a check on the parameter does not pass.
        """

        # parse args
        feature_series = args[0]
        target_series = args[1]

        raise_if_not(all([s.has_same_time_as(target_series) for s in feature_series]), 'All provided time series must '
                     'have the same time index', logger)

        _backtest_general_checks(target_series, kwargs)

    def _backtest_model_specific_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Method to be overriden in subclass for model specific sanity checks"""
        pass

    @_with_sanity_checks("_backtest_sanity_checks", "_backtest_model_specific_sanity_checks")
    def backtest(self,
                 feature_series: Iterable[TimeSeries],
                 target_series: TimeSeries,
                 start: Union[pd.Timestamp, float, int] = 0.7,
                 forecast_horizon: int = 1,
                 stride: int = 1,
                 trim_to_series: bool = True,
                 verbose=False) -> TimeSeries:
        """ A function for backtesting `RegressionModel`'s.

        This function computes the time series of historical predictions
        that would have been obtained, if the current model had been used to predict `target_series`
        using the `feature_series`, with a certain time horizon.

        To this end, it repeatedly builds a training set composed of both features and targets,
        from `feature_series` and `target_series`, respectively.
        It trains the current model on the training set, emits a (point) prediction for a fixed
        forecast horizon, and then moves the end of the training set forward by `stride`
        time steps. The resulting predictions are then returned.

        This always re-trains the models on the entire available history,
        corresponding an expending window strategy.

        Parameters
        ----------
        feature_series
            A list of time series representing the features for the regression model (independent variables)
        target_series
            The univariate target time series for the regression model (dependent variable)
        start
            The first prediction time, at which a prediction is computed for a future time
        forecast_horizon
            The forecast horizon for the point predictions
        stride
            The number of time steps (the unit being the frequency of `series`) between two consecutive predictions.
        trim_to_series
            Whether the predicted series has the end trimmed to match the end of the main series
        verbose
            Whether to print progress

        Returns
        -------
        TimeSeries
            A time series containing the forecast values when successively applying
            the current model with the specified forecast horizon.
        """
        start = _get_timestamp_at_point(start, target_series)

        # build the prediction times in advance (to be able to use tqdm)
        if trim_to_series:
            last_pred_time = target_series.time_index()[-forecast_horizon - stride]
        else:
            last_pred_time = target_series.time_index()[-stride - 1]

        # build the prediction times in advance (to be able to use tqdm)
        pred_times = [start]
        while pred_times[-1] <= last_pred_time:
            pred_times.append(pred_times[-1] + target_series.freq() * stride)

        # what we'll return
        values = []
        times = []

        iterator = _build_tqdm_iterator(pred_times, verbose)

        for pred_time in iterator:
            # build train/val series
            train_features = [s.drop_after(pred_time) for s in feature_series]
            train_target = target_series.drop_after(pred_time)
            val_features = [s.slice_n_points_after(pred_time, forecast_horizon) for s in feature_series]

            self.fit(train_features, train_target)
            pred = self.predict(val_features)
            values.append(pred.values()[-1])  # store the N-th point
            times.append(pred.end_time())  # store the N-th timestamp

        return TimeSeries.from_times_and_values(pd.DatetimeIndex(times), np.array(values))

    def residuals(self) -> TimeSeries:
        """ Computes the time series of residuals of this model on the training time series

        The residuals are computed as

        .. math:: z_t := y_t - \\hat{y}_t,

        where :math:`y_t` is the actual target time series over the training set,
        and :math:`\\hat{y}_t` is the time series of predicted targets, over the training set.

        Returns
        -------
        TimeSeries
            The time series containing the residuals
        """

        if (not self._fit_called):
            raise_log(Exception('fit() must be called before predict()'), logger)

        train_pred = self.predict(self.train_features)
        return self.train_target - train_pred



"""
Scaler wrapper
--------------
"""
from ..timeseries import TimeSeries
from ..logging import get_logger, raise_log
from sklearn.preprocessing import MinMaxScaler

logger = get_logger(__name__)


class ScalerWrapper:
    def __init__(self, scaler=MinMaxScaler(feature_range=(0, 1))):
        """
        Generic wrapper class for using transformers/scalers that implement `fit()`, `transform()` and
        `inverse_transform()` methods (typically from scikit-learn) on `TimeSeries`.

        Parameters
        ----------
        scaler
            The transformer/scaler to transform the data.
            It must provide the `fit()`, `transform()` and `inverse_transform()` methods.
            Default: `sklearn.preprocessing.MinMaxScaler(feature_range=(0, 1))`; this
            will scale all the values of a time series between 0 and 1.
        """
        def _raise():
            raise_log(ValueError(
                      'The provided transformer object must have fit(), transform() and inverse_transform() methods'),
                      logger)

        if (not callable(getattr(scaler, "fit", None)) or not callable(getattr(scaler, "transform", None))
                or not callable(getattr(scaler, "inverse_transform", None))): # noqa W503 
            _raise()

        self.transformer = scaler
        self.train_series = None
        self._fit_called = False

    def fit(self, series: TimeSeries) -> 'ScalerWrapper':
        """ Fits this transformer/scaler to the provided time series data

        Parameters
        ----------
        series
            The time series to fit the transformer on
        """
        self.transformer.fit(series.values().reshape((-1, series.width)))
        self.train_series = series
        self._fit_called = True
        return self

    def transform(self, series: TimeSeries) -> TimeSeries:
        """
        Returns a new time series, transformed with this (fitted) transformer.
        This does not handle series with confidence intervals - the intervals are discarded.

        Parameters
        ----------
        series
            The time series to transform

        Returns
        -------
        TimeSeries
            A new time series, transformed with this (fitted) transformer.
        """
        assert self._fit_called, 'fit() must be called before transform()'
        return TimeSeries.from_times_and_values(series.time_index(),
                                                self.transformer.transform(series.values().
                                                                           reshape((-1, series.width))),
                                                series.freq())

    def fit_transform(self, series: TimeSeries) -> TimeSeries:
        """
        Applies `fit()` and `transform()` in one go.

        Parameters
        ----------
        series
            The time series to fit the transformer and to transform
        Returns
        -------
            A new time series, transformed with this (fitted) transformer.
        """
        return self.fit(series).transform(series)

    def inverse_transform(self, series: TimeSeries) -> TimeSeries:
        """
        Performs the inverse transformation on a time series

        Parameters
        ----------
        series
            The time series to inverse transform

        Returns
        -------
        TimeSeries
            The inverse transform
        """
        return TimeSeries.from_times_and_values(series.time_index(),
                                                self.transformer.inverse_transform(series.values().
                                                                                   reshape((-1, series.width))),
                                                series.freq())



import shutil
import unittest
import logging
from ..utils import timeseries_generation as tg
from ..logging import get_logger

logger = get_logger(__name__)

try:
    import torch
    from ..models.tcn_model import TCNModel
    from .test_RNN import RNNModelTestCase
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch not available. TCN tests will be skipped.')
    TORCH_AVAILABLE = False


if TORCH_AVAILABLE:
    class TCNModelTestCase(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            logging.disable(logging.CRITICAL)

        @classmethod
        def tearDownClass(cls):
            shutil.rmtree('.darts')

        def test_creation(self):
            with self.assertRaises(ValueError):
                # cannot choose a kernel size larger than the input length
                TCNModel(kernel_size=100, input_length=20)
            TCNModel()

        def test_fit(self):
            large_ts = tg.constant_timeseries(length=100, value=1000)
            small_ts = tg.constant_timeseries(length=100, value=10)

            # Test basic fit and predict
            model = TCNModel(n_epochs=20, num_layers=1)
            model.fit(large_ts[:98])
            pred = model.predict(n=2).values()[0]

            # Test whether model trained on one series is better than one trained on another
            model2 = TCNModel(n_epochs=20, num_layers=1)
            model2.fit(small_ts[:98])
            pred2 = model2.predict(n=2).values()[0]
            self.assertTrue(abs(pred2 - 10) < abs(pred - 10))

            # test short predict
            pred3 = model2.predict(n=1)
            self.assertEqual(len(pred3), 1)

        def test_coverage(self):
            torch.manual_seed(0)
            input_lengths = range(20, 50)
            kernel_sizes = range(2, 5)
            dilation_bases = range(2, 5)

            for kernel_size in kernel_sizes:
                for dilation_base in dilation_bases:
                    if dilation_base > kernel_size:
                        continue
                    for input_length in input_lengths:

                        # create model with all weights set to one
                        model = TCNModel(kernel_size=kernel_size, dilation_base=dilation_base, input_length=input_length,
                                         weight_norm=False)
                        for res_block in model.model.res_blocks:
                            res_block.conv1.weight = torch.nn.Parameter(torch.ones(res_block.conv1.weight.shape))
                            res_block.conv2.weight = torch.nn.Parameter(torch.ones(res_block.conv2.weight.shape))

                        model.model.eval()
                        input_tensor = torch.zeros([1, input_length, 1], dtype=torch.float)
                        zero_output = model.model.forward(input_tensor).float()[0, -1, 0]

                        # test for full coverage
                        for i in range(input_length):
                            input_tensor[0, i, 0] = 1
                            curr_output = model.model.forward(input_tensor).float()[0, -1, 0]
                            self.assertNotEqual(zero_output, curr_output)
                            input_tensor[0, i, 0] = 0

                        # create model with all weights set to one and one layer less than is automatically detected
                        model_2 = TCNModel(kernel_size=kernel_size, dilation_base=dilation_base, input_length=input_length,
                                           weight_norm=False, num_layers=model.model.num_layers - 1)
                        for res_block in model_2.model.res_blocks:
                            res_block.conv1.weight = torch.nn.Parameter(torch.ones(res_block.conv1.weight.shape))
                            res_block.conv2.weight = torch.nn.Parameter(torch.ones(res_block.conv2.weight.shape))

                        model_2.model.eval()
                        input_tensor = torch.zeros([1, input_length, 1], dtype=torch.float)
                        zero_output = model_2.model.forward(input_tensor).float()[0, -1, 0]

                        # test for incomplete coverage
                        uncovered_input_found = False
                        if model_2.model.num_layers == 1:
                            continue
                        for i in range(input_length):
                            input_tensor[0, i, 0] = 1
                            curr_output = model_2.model.forward(input_tensor).float()[0, -1, 0]
                            if (zero_output == curr_output):
                                uncovered_input_found = True
                                break
                            input_tensor[0, i, 0] = 0
                        self.assertTrue(uncovered_input_found)

        def test_use_full_output_length(self):
            series = tg.linear_timeseries(length=100)
            RNNModelTestCase.helper_test_use_full_output_length(self, TCNModel, series)

        def test_multivariate(self):
            series_multivariate = tg.linear_timeseries(length=100).stack(tg.linear_timeseries(length=100))
            RNNModelTestCase.helper_test_multivariate(self, TCNModel, series_multivariate)
